local deformat = AceLibrary("Deformat-2.0")
local tablet = AceLibrary("Tablet-2.0")
local roster = AceLibrary("Roster-2.1")
local _G = getfenv(0)

-- Reference to our attendance tracker details frame, also used to show lists of
-- special items.
local trackerFrame = nil

-- Functions
local setFrameHeader = nil
local setText = nil
local createTrackerFrame = nil

-- State variables
local currentTrackset = nil
local grouped = nil
local lastBossKilled = nil

-- These items are always ignored by the loot handler.
local ignoredItems = {
	[30316] = true, -- Devastation
	[30317] = true, -- Cosmic Infuser
	[30312] = true, -- Infinity Blade
	[30318] = true, -- Netherstrand Longbow
	[30313] = true, -- Staff of Disintegration
	[30311] = true, -- Warp Slicer
	[30320] = true, -- Bundle of Nether Spikes
	[30314] = true, -- Phaseshift Bulwark
	[30319] = true, -- Nether Spikes
	[29434] = true, -- Badge of Justice
	[22450] = true, -- Void Crystal
	[20725] = true, -- Nexus Crystal
}

-- These items are tracked in a "Special Items" menu, regardless of their
-- quality. Items listed here will never be added to the standard "looted by"
-- list.
local specialItems = {
	[32897] = true, -- Mark of the Illidari
	[32428] = true, -- Heart of Darkness

	-- Epic gems
	[32230] = true, -- Shadowsong Amethyst
	[32249] = true, -- Seaspray Emerald
	[32231] = true, -- Pyrestone
	[32229] = true, -- Lionseye
	[32228] = true, -- Empyrean Sapphire
	[32227] = true, -- Crimson Spinel
}

local outputFormatters = {
	-- The different types of output formatters available.
}
local formatters = {
	-- Keyed lookup table holding the references to the formatters.
}

local killedBosses = {
	-- Lists the bosses killed since session start, not counting ignored bosses.
}

local worldbosses = {
	-- Holds a map of the worldboss class mobs we've seen since login.
}

local ignoredBosses = {
	["Krosh Firehand"] = true,
	["Olm the Summoner"] = true,
	["Kiggler the Crazed"] = true,
	["Blindeye the Seer"] = true,
	["Thaladred the Darkener"] = true,
	["Master Engineer Telonicus"] = true,
	["Grand Astromancer Capernian"] = true,
	["Lord Sanguinar"] = true,
	["Hellfire Channeler"] = true,
	["Amani Bear Spirit"] = true,
	["Amani Lynx Spirit"] = true,
	["Amani Eagle Spirit"] = true,
}

-- Map of bosses we ignore by default.
if AceLibrary:HasInstance("Babble-Boss-2.2") then
	local bb = AceLibrary("Babble-Boss-2.2")
	for k,v in pairs(ignoredBosses) do
		if bb:HasTranslation(k) then
			ignoredBosses[k] = nil
			ignoredBosses[bb[k]] = true
		end
	end
end

-- Helper table to cache hex colors for all classes.
local hexColors = setmetatable({}, {__index =
	function(self, key)
		local r, g, b = 0.63, 0.63, 0.63
		if RAID_CLASS_COLORS and RAID_CLASS_COLORS[key] then
			r, g, b = RAID_CLASS_COLORS[key].r, RAID_CLASS_COLORS[key].g, RAID_CLASS_COLORS[key].b
		end
		self[key] = "|cff" .. string.format("%02x%02x%02x", r * 255, g * 255, b * 255)
		return self[key]
	end
})

-- Helper table to cache colored player names.
local coloredName = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local class = roster.roster[key] and roster.roster[key].class or "WTF"
		self[key] = hexColors[class] .. key .. "|r"
		return self[key]
	end
})

-- We'll worry about locales when someone wants to translate it :)
local L = setmetatable({}, {__index =
	function(self, key)
		self[key] = key
		return key
	end
})

_G.NRT = AceLibrary("AceAddon-2.0"):new(
	"AceEvent-2.0",
	"AceDB-2.0",
	"AceConsole-2.0",
	"FuBarPlugin-2.0",
	"AceHook-2.1"
)
local addon = _G.NRT
addon.revision = tonumber(("$Revision: 59989 $"):sub(12, -3))
addon.version = (addon.version or "3") .. addon.revision

local function get(key)
	return addon.db.profile[key]
end
local function set(key, val)
	addon.db.profile[key] = val
end

local options
options = {
	type = "group",
	args = {
		attendance = {
			type = "header",
			name = L["Attendance"],
			order = 100,
		},
		announce = {
			type = "group",
			name = L["Announce"],
			desc = L["Options for attendance announcements."],
			order = 101,
			args = {
				enable = {
					type = "toggle",
					name = L["Enable"],
					desc = L["Enable or disable auto-announcing when taking attendance."],
					get = get,
					set = set,
					passValue = "announceAttendance",
					order = 1,
				},
				guild = {
					type = "toggle",
					name = L["Guild"],
					desc = L["Announce when taking attendance to the guild."],
					get = get,
					set = set,
					passValue = "guildAttendance",
					order = 2,
					disabled = function() return not addon.db.profile.announceAttendance end,
				},
				custom = {
					type = "text",
					name = L["Custom channel"],
					desc = L["Announce when taking attendance to a custom channel as well."],
					usage = L["<channelName or blank to not announce>"],
					get = function()
						return addon.db.profile.customAttendance
					end,
					set = function(v)
						if type(v) == "string" then v = v:trim() end
						if type(v) == "string" and v:len() > 0 then
							addon.db.profile.customAttendance = v
						else
							addon.db.profile.customAttendance = nil
						end
					end,
					order = 3,
					disabled = function() return not addon.db.profile.announceAttendance end,
				},
			},
		},
		announceTimeout = {
			type = "range",
			name = L["Timeout"],
			desc = L["How long, in minutes, the timeout should be from when you take attendance until people can no longer whisper you."],
			order = 102,
			min = 1,
			max = 10,
			step = 1,
			get = get,
			set = set,
			passValue = "announceTimeout",
			disabled = function() return not addon.db.profile.announceAttendance end,
		},
		add = {
			type = "text",
			name = L["Add player"],
			desc = L["Adds a player to the latest attendance set."],
			order = 103,
			get = false,
			usage = L["<player name>"],
			set = function(newPlayer)
				newPlayer = newPlayer:trim():lower()
				local size = #addon.db.profile.tracked
				local set = addon.db.profile.tracked[size]
				if set then
					local found = nil
					for i, v in ipairs(set.attendants) do
						if v:lower() == newPlayer then
							found = true
							break
						end
					end
					if not found then
						table.insert(set.attendants, newPlayer)
					end
				end
			end,
			disabled = function() return type(addon.db.profile.tracked) ~= "table" or #addon.db.profile.tracked == 0 end,
		},
		whitelist = {
			type = "group",
			name = L["Whitelist"],
			desc = L["Players added to the whitelist will be accepted on the attendance sets without being in your guild."],
			order = 104,
			args = {
				spacer = {
					type = "header",
					name = " ",
					order = 50,
					hidden = function() return not next(addon.db.profile.whitelist) end
				},
				add = {
					type = "text",
					name = L["Add"],
					desc = L["Type in the name of the player to whitelist. The name is case-insensitive."],
					usage = L["<player name>"],
					get = false,
					set = function(newPlayer)
						if type(newPlayer) == "string" then
							newPlayer = newPlayer:trim():lower()
							if newPlayer:len() > 2 and newPlayer ~= "add" and newPlayer ~= "spacer" then
								addon.db.profile.whitelist[newPlayer] = true
								addon:UpdateWhitelist()
							else
								addon:Print(L["Either the name is less than 3 characters long, or you tried to add the words 'add' or 'spacer', which are reserved, you can't whitelist them."])
							end
						end
					end,
					order = 100,
				},
			},
		},
		groups = {
			type = "range",
			name = L["Groups"],
			desc = L["Only include people in the groups up to the number set here. Defaults to 5, which means that only people in group 1-5 will be part of the default attendance snapshot, so you can use groups 6-8 as waiting list groups."],
			order = 105,
			min = 1,
			max = 8,
			step = 1,
			get = get,
			set = set,
			passValue = "groups",
		},
		boss = {
			type = "toggle",
			name = L["Take attendance on boss kill"],
			desc = L["Automatically pop up a dialog asking if you want to take attendance now when a boss is killed."],
			order = 106,
			get = get,
			set = set,
			passValue = "bossKill",
		},
		spacer1 = {
			type = "header",
			name = " ",
			order = 150,
		},
		other = {
			type = "header",
			name = L["Other options"],
			order = 200,
		},
		announceLoot = {
			type = "toggle",
			name = L["Announce loot to guild"],
			desc = L["Automatically announce to guild chat when someone loots an epic item that is not ignored."],
			order = 201,
			get = get,
			set = set,
			passValue = "announceLoot",
		},
		special = {
			type = "group",
			name = L["Special items"],
			desc = L["Lists the special items looted during raids."],
			order = 202,
			args = {},
			disabled = function()
				return not next(options.args.special.args)
			end,
		},
		formatter = {
			type = "text",
			name = L["Output format"],
			desc = L["Choose the output format for attendance details."],
			order = 203,
			get = get,
			set = set,
			passValue = "formatter",
			validate = outputFormatters,
		},
		spacer2 = {
			type = "header",
			name = " ",
			order = 250,
		},
		clearAll = {
			type = "execute",
			name = L["Clear all"],
			desc = L["Clears all the recorded information, including loot, special items and attendance sets."],
			order = 400,
			func = "ClearAll",
		},
	},
}

addon.hasIcon = "Interface\\Icons\\Spell_Holy_DivineIntervention"

-----------------------------------------------------------------------
-- Initialization

function addon:OnInitialize()
	self.clickableTooltip = true
	self.hasNoColor = true
	self.hideMenuTitle = true

	self:RegisterDB("NRTDB")
	self:RegisterDefaults("profile", {
		loot = {},
		tracked = {},
		specialItems = {},
		customAttendance = "nstandby",
		guildAttendance = true,
		announceAttendance = true,
		bossKill = true,
		announceLoot = true,
		whitelist = {},
		trackerFrameWidth = 280,
		trackerFrameHeight = 420,
		formatter = "Nihilum",
		dbFormat = 1,
		announceTimeout = 2,
		groups = 5,
		itemNotes = {},
	})

	self:UpdateWhitelist()
	self:UpdateSpecialItemList()

	self.OnMenuRequest = options

	self:RegisterChatCommand("/nrt", options, "NRT")


	-- Convert old date/time pairs to new format, unix time.
	-- Can probably remove this code after a while.
	if self.db.profile.dbFormat == 1 then
		local year = tonumber(date("%Y"))
		for i, v in ipairs(self.db.profile.loot) do
			if type(v.date) == "string" then
				local day, month, hour, minute = select(3, v.date:find("(%d+)/(%d+) (%d+):(%d+)"))
				local newDate = time({
					year = year,
					month = tonumber(month),
					day = tonumber(day),
					hour = tonumber(hour),
					min = tonumber(minute)
				})
				v.date = newDate
			end
		end
		for i, v in ipairs(self.db.profile.tracked) do
			if type(v.date) == "string" then
				local day, month, hour, minute = select(3, v.date:find("(%d+)/(%d+) (%d+):(%d+)"))
				local newDate = time({
					year = year,
					month = tonumber(month),
					day = tonumber(day),
					hour = tonumber(hour),
					min = tonumber(minute)
				})
				v.date = newDate
			end
		end
		self.db.profile.dbFormat = 2
	end
end

function addon:OnEnable(first)
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end
	if type(popup["NRTBossKilled"]) ~= "table" then
		popup["NRTBossKilled"] = {
			text = "",
			button1 = TEXT(YES),
			button2 = TEXT(NO),
			timeout = 0,
			whileDead = 1,
			hideOnEscape = 1,
			OnAccept = function() self:TakeAttendance() end
		}
	end

	self:RegisterEvent("RosterLib_RosterUpdated")
end

-----------------------------------------------------------------------
-- Dynamic menu updating

local function removeWhitelistedUnit(unit)
	addon.db.profile.whitelist[unit] = nil
	options.args.whitelist.args[unit] = nil
end

function addon:UpdateWhitelist()
	local opt = options.args.whitelist.args
	local db = self.db.profile.whitelist
	for k in pairs(db) do
		if not opt[k] then
			opt[k] = {
				type = "execute",
				name = k,
				desc = L["This player is allowed on the attendance list, click to remove."],
				func = removeWhitelistedUnit,
				passValue = k,
				order = 10,
			}
		end
	end
end

local function showSpecialItems(item)
	createTrackerFrame()
	local db = addon.db.profile.specialItems[item]
	if not db then return end
	local text = ""
	for k, v in pairs(db) do
		-- I know this isn't ideal, but what the hell, it happens rarely.
		text = text .. tostring(v) .. "   " .. k .. "\n"
	end
	setText(text)
	local name = GetItemInfo(item)
	setFrameHeader(name or item)
	trackerFrame.trackingId = -1
	trackerFrame:Show()
end

local function clearSpecialItem(item)
	addon.db.profile.specialItems[item] = nil
	addon:UpdateSpecialItemList()
end

-- I'd like to not re-create the entire menu every time, but it's not something
-- I prioritize.
function addon:UpdateSpecialItemList()
	local opt = options.args.special.args
	local db = self.db.profile.specialItems
	for k in pairs(opt) do opt[k] = nil end
	for k in pairs(db) do
		local _, name = GetItemInfo(k)
		if not name then name = "ID: " .. tostring(k) end
		opt[k] = {
			type = "execute",
			name = name,
			desc = L["Show who looted %s and how many they looted."]:format(name),
			func = showSpecialItems,
			passValue = k,
			order = 10,
		}
		opt[k .. "flush"] = {
			type = "execute",
			name = L["Flush %s"]:format(name),
			desc = L["Removes all the stores loot information on %s from the database."]:format(name),
			func = clearSpecialItem,
			passValue = k,
			order = 100,
		}
	end
	if next(opt) then
		opt.spacer = {
			type = "header",
			name = " ",
			order = 50,
		}
	end
end

-----------------------------------------------------------------------
-- Attendance tracking

function addon:TakeAttendance()
	if self:IsEventScheduled("Timeout") then
		self:Print(L["You're already taking attendance, please wait."])
		return
	end

	local name = nil
	if lastBossKilled then
		name = lastBossKilled
	else
		name = GetRealZoneText()
	end
	local newTrackset = {
		name = name,
		zone = GetRealZoneText(),
		attendants = {},
		date = time(),
	}
	for n, u in pairs(roster.roster) do
		if u.class ~= "PET" and u.subgroup <= self.db.profile.groups then
			table.insert(newTrackset.attendants, n)
		end
	end
	table.insert(self.db.profile.tracked, newTrackset)
	currentTrackset = #self.db.profile.tracked

	if self.db.profile.announceAttendance then
		self:Hook("ChatFrame_MessageEventHandler", "WhisperHandler", true)
		local str = nil
		if lastBossKilled then
			str = L["NRT: %s down! Whisper me your mains name to be added to the DKP list."]:format(lastBossKilled)
		else
			str = L["NRT: Whisper me your mains name to be added to the DKP list."]
		end

		if self.db.profile.guildAttendance then
			SendChatMessage(("*"):rep(10), "GUILD")
			SendChatMessage(str, "GUILD")
			SendChatMessage(("*"):rep(10), "GUILD")
		end

		if self.db.profile.customAttendance then
			local id = GetChannelName(self.db.profile.customAttendance)
			if id then
				SendChatMessage(("*"):rep(10), "CHANNEL", nil, id)
				SendChatMessage(str, "CHANNEL", nil, id)
				SendChatMessage(("*"):rep(10), "CHANNEL", nil, id)
			end
		end

		self:ScheduleEvent("Timeout", self.AttendanceTimeout, self.db.profile.announceTimeout * 60, self)

		if type(ShutUp_Disable) == "function" then
			ShutUp_Disable()
		end
	else -- We're not announcing anything, so just clean up right away.
		self:AttendanceTimeout()
	end
end

function createTrackerFrame()
	if trackerFrame then return end

	trackerFrame = CreateFrame("Frame", "NRTTrackerFrame", UIParent)
	trackerFrame:Hide()

	trackerFrame:SetWidth(addon.db.profile.trackerFrameWidth)
	trackerFrame:SetHeight(addon.db.profile.trackerFrameHeight)

	-- thanks to haste for the style
	trackerFrame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\AddOns\\NRT\\textures\\otravi-semi-full-border", edgeSize = 32,
		insets = {left = 1, right = 1, top = 20, bottom = 1},
	})
	trackerFrame:SetBackdropColor(24/255, 24/255, 24/255)
	trackerFrame:ClearAllPoints()
	trackerFrame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	trackerFrame:EnableMouse(true)
	trackerFrame:RegisterForDrag("LeftButton")
	trackerFrame:SetMovable(true)
	trackerFrame:SetResizable(true)
	trackerFrame:SetMinResize(280,300)
	trackerFrame:SetScript("OnDragStart", function()
		this:StartMoving()
	end)
	trackerFrame:SetScript("OnDragStop", function()
		this:StopMovingOrSizing()
		local s = this:GetEffectiveScale()

		addon.db.profile.posX = this:GetLeft() * s
		addon.db.profile.posY = this:GetTop() * s
	end)
	trackerFrame:SetScript("OnSizeChanged", function()
		addon.db.profile.trackerFrameWidth = this:GetWidth()
		addon.db.profile.trackerFrameHeight = this:GetHeight()
	end)

	local cheader = trackerFrame:CreateFontString(nil,"OVERLAY")
	cheader:ClearAllPoints()
	cheader:SetWidth(240)
	cheader:SetHeight(15)
	cheader:SetPoint("TOPLEFT", trackerFrame, "TOPLEFT", 7, -14)
	cheader:SetPoint("TOPRIGHT", trackerFrame, "TOPRIGHT", -28, -14)
	cheader:SetFont("Fonts\\FRIZQT__.TTF", 12)
	cheader:SetJustifyH("LEFT")
	setFrameHeader = function(text)
		cheader:SetText(text)
	end
	cheader:SetText("Tracker")
	cheader:SetShadowOffset(.8, -.8)
	cheader:SetShadowColor(0, 0, 0, 1)

	local close = trackerFrame:CreateTexture(nil, "ARTWORK")
	close:SetTexture("Interface\\AddOns\\NRT\\textures\\otravi-close")
	close:SetTexCoord(0, .625, 0, .9333)
	close:SetWidth(20)
	close:SetHeight(14)
	close:SetPoint("TOPRIGHT", trackerFrame, "TOPRIGHT", -7, -15)

	local closebutton = CreateFrame("Button", nil)
	closebutton:SetParent( trackerFrame )
	closebutton:SetWidth(20)
	closebutton:SetHeight(14)
	closebutton:SetPoint("CENTER", close, "CENTER")
	closebutton:SetScript( "OnClick", function()
		trackerFrame:Hide()
		trackerFrame.trackingId = nil
	end)

	local scroll = CreateFrame("ScrollFrame", "NRTScrollFrame", trackerFrame, "UIPanelScrollFrameTemplate")
	scroll:SetPoint("TOPLEFT", trackerFrame, "TOPLEFT", 20, -40)
	scroll:SetPoint("RIGHT", trackerFrame, "RIGHT", -40, 0)
	scroll:SetPoint("BOTTOM", trackerFrame, 0, 20)

	local textField = CreateFrame("EditBox", "NRTEditBox", scroll)
	textField:SetFontObject(ChatFontNormal)
	textField:SetMultiLine(true)
	textField:SetWidth(240)
	textField:SetScript("OnEscapePressed", function() this:ClearFocus() end)
	textField:SetScript("OnTextChanged", function() scroll:UpdateScrollChildRect() end)
	textField:SetText("")
	setText = function(text)
		textField:SetText(text)
	end
	textField:SetAutoFocus(false)
	scroll:SetScrollChild(textField)
	
	-- drag handle, thanks to ammo
	local draghandle = CreateFrame("Frame", nil, trackerFrame)
	draghandle:Hide()
	draghandle:SetFrameLevel( trackerFrame:GetFrameLevel() + 10 )
	draghandle:SetWidth(16)
	draghandle:SetHeight(16)
	draghandle:SetPoint("BOTTOMRIGHT", trackerFrame, "BOTTOMRIGHT", -1, 1)
	draghandle:EnableMouse(true)
	draghandle:SetScript("OnMouseDown", function()
		trackerFrame:StartSizing("BOTTOMRIGHT")
	end)
	draghandle:SetScript("OnMouseUp", function()
		trackerFrame:StopMovingOrSizing()
	end)
	draghandle:SetScript("OnEnter", function()
		addon:CancelScheduledEvent("NRTHideDrag")
	end)
	draghandle:SetScript("OnLeave", function()
		this:Hide()
	end)

	local texture = draghandle:CreateTexture(nil,"BACKGROUND")
	texture:SetTexture("Interface\\AddOns\\NRT\\textures\\draghandle")
	texture:SetWidth(16)
	texture:SetHeight(16)
	texture:SetBlendMode("ADD")
	texture:SetPoint("CENTER", draghandle, "CENTER", 0, 0)

	trackerFrame:SetScript("OnEnter", function()
		draghandle:Show()
	end)
	local function hideHandle()
		draghandle:Hide()
	end
	trackerFrame:SetScript("OnLeave", function()
		addon:ScheduleEvent("NRTHideDrag", hideHandle, 1)
	end)
	
	local x = addon.db.profile.posX
	local y = addon.db.profile.posY
	if x and y then
		local s = trackerFrame:GetEffectiveScale()
		trackerFrame:ClearAllPoints()
		trackerFrame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x / s, y / s)
	end
end

-----------------------------------------------------------------------
-- Event handling

local function checkTarget(target)
	local class = UnitClassification(target)
	if class and class == "worldboss" then
		worldbosses[UnitName(target)] = true
	end
end

function addon:PLAYER_TARGET_CHANGED()
	checkTarget("target")
end

function addon:UPDATE_MOUSEOVER_UNIT()
	checkTarget("mouseover")
end

function addon:AttendanceTimeout()
	if self:IsHooked("ChatFrame_MessageEventHandler") then
		self:Unhook("ChatFrame_MessageEventHandler")
	end

	if self.db.profile.announceAttendance then
		local str = L["NRT: If you haven't whispered me by now, you're too late."]
		if self.db.profile.guildAttendance then
			SendChatMessage(str, "GUILD")
		end
		if self.db.profile.customAttendance then
			local id = GetChannelName(self.db.profile.customAttendance)
			if id then
				SendChatMessage(str, "CHANNEL", nil, id)
			end
		end
		if type(ShutUp_Enable) == "function" then
			ShutUp_Enable()
		end
	end
	currentTrackset = nil
end

function addon:RosterLib_RosterUpdated()
	local num = GetNumRaidMembers()
	if not grouped and num > 0 then
		grouped = true
		self:RegisterEvent("CHAT_MSG_COMBAT_HOSTILE_DEATH")
		self:RegisterEvent("CHAT_MSG_LOOT")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	elseif grouped and num == 0 then
		grouped = nil
		self:UnregisterEvent("CHAT_MSG_COMBAT_HOSTILE_DEATH")
		self:UnregisterEvent("CHAT_MSG_LOOT")
		self:UnregisterEvent("PLAYER_TARGET_CHANGED")
		self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
	end
end

function addon:CHAT_MSG_LOOT(msg)
	if select(2, IsInInstance()) ~= "raid" then return end

	local player, item = deformat(msg, LOOT_ITEM)
	if not player then
		item = deformat(msg, LOOT_ITEM_SELF)
		if item then
			player = UnitName("player")
		end
	end

	if type(item) == "string" and type(player) == "string" then
		local itemName, itemLink, itemRarity = GetItemInfo(item)

		-- Find the item ID to make sure it's not a ignored item
		local itemId = select(3, itemLink:find("item:(%d+):"))
		if not itemId then return end
		itemId = tonumber(itemId:trim())
		if type(itemId) ~= "number" then return end
		if ignoredItems[itemId] or (itemRarity < 4 and not specialItems[itemId]) then return end

		if specialItems[itemId] then
			-- Figure out how many they looted.
			local amount
			if player == UnitName("player") then
				amount = select(2, deformat(msg, LOOT_ITEM_SELF_MULTIPLE))
			else
				amount = select(3, deformat(msg, LOOT_ITEM_MULTIPLE))
			end
			amount = tonumber(amount)
			if type(amount) ~= "number" then amount = 1 end
			if not self.db.profile.specialItems[itemId] then self.db.profile.specialItems[itemId] = {} end
			if not self.db.profile.specialItems[itemId][player] then self.db.profile.specialItems[itemId][player] = 0 end
			self.db.profile.specialItems[itemId][player] = self.db.profile.specialItems[itemId][player] + amount
			self:UpdateSpecialItemList()
		else
			local unit = roster:GetUnitObjectFromName(player)
			if not unit then return end

			local lootedFrom = nil
			local targetId = unit.unitid .. "target"
			if UnitExists(targetId) and (UnitClassification(targetId) == "worldboss" or UnitClassification(targetId) == "elite") then
				lootedFrom = UnitName(targetId)
			elseif lastBossKilled then
				lootedFrom = lastBossKilled
			else
				lootedFrom = GetRealZoneText()
			end

			if self.db.profile.announceLoot then
				SendChatMessage(L["%s looted %s (%s)!"]:format(player, itemLink, lootedFrom), "GUILD")
			end

			table.insert(self.db.profile.loot, {
				player = player,
				item = itemLink,
				from = lootedFrom,
				date = time(),
			})
		end
	end
end

local function ResetBossKilled()
	lastBossKilled = nil
end

function addon:CHAT_MSG_COMBAT_HOSTILE_DEATH(msg)
	if select(2, IsInInstance()) ~= "raid" then return end

	local mob = deformat(msg, UNITDIESOTHER)
	if mob and worldbosses[mob] and not ignoredBosses[mob] then
		table.insert(killedBosses, mob)

		lastBossKilled = mob
		self:ScheduleEvent("ResetBossKilled", ResetBossKilled, 300)

		if self.db.profile.bossKill then
			_G.StaticPopupDialogs["NRTBossKilled"].text = L["%s killed. Take raid attendance now?"]:format(mob)
			_G.StaticPopup_Show("NRTBossKilled")
		end
	end
end

-----------------------------------------------------------------------
-- Hooks

local guildMemberList = nil
local function IsGuildMember(name)
	if not guildMemberList then
		guildMemberList = {}

		local offline = GetGuildRosterShowOffline()
		local selection = GetGuildRosterSelection()
		SetGuildRosterShowOffline(true)
		SetGuildRosterSelection(0)
		GetGuildRosterInfo(0)
		local numGuildMembers = GetNumGuildMembers()
		for i = 1, numGuildMembers, 1 do
			local n = GetGuildRosterInfo(i)
			if n then
				guildMemberList[n:lower()] = true
			end
		end
		SetGuildRosterShowOffline(offline)
		SetGuildRosterSelection(selection)
	end
	return guildMemberList[name]
end

local function IsGroupMember(name)
	local found = nil
	for n, u in pairs(roster.roster) do
		if u.class ~= "PET" and n:lower() == name and u.subgroup <= addon.db.profile.groups then
			found = true
			break
		end
	end
	return found
end

function addon:WhisperHandler(event, ...)
	if event == "CHAT_MSG_WHISPER" and type(arg1) == "string" and type(arg2) == "string" then
		local player = arg2
		local text = arg1:lower():trim()
		if not text:find("%s") and not text:find("%d") then
			if not IsGroupMember(text) and (self.db.profile.whitelist[text] or IsGuildMember(text)) then
				local set = self.db.profile.tracked[currentTrackset]
				if set then
					local found = nil
					for i, v in ipairs(set.attendants) do
						if v == text or v:lower() == text then
							found = true
							break
						end
					end
					if not found then
						table.insert(set.attendants, text)
						SendChatMessage(L["<NRT>: %q has been added to the attendance list."]:format(text), "WHISPER", nil, player)
						self:Print(L["%q has been added to the attendance list by %s."]:format(coloredName[text], coloredName[player]))
						if trackerFrame and trackerFrame:IsShown() and trackerFrame.trackingId == currentTrackset then
							self:ShowTrackerForRaidIndex(currentTrackset)
						end
					end
				end
			else
				SendChatMessage(L["<NRT>: %q is not in the guild, group or whitelist."]:format(text), "WHISPER", nil, player)
			end
			return
		end
	elseif event == "CHAT_MSG_WHISPER_INFORM" and type(arg1) == "string" and arg1:find("^<NRT>: ") then
		return
	end
	return self.hooks["ChatFrame_MessageEventHandler"](event, ...)
end

-----------------------------------------------------------------------
-- UI Handling

function addon:ClearAll()
	self.db.profile.loot = {}
	self.db.profile.tracked = {}
	self.db.profile.specialItems = {}

	self:UpdateSpecialItemList()
	self:UpdateDisplay()
end

local currentNoteItem = nil
local itemNoteText = L["Set note for %s.\n\nSetting a note sets it for all instances - past, present and future - of this item, not just this particular one."]
local function itemClickFunc(item, index)
	if type(index) ~= "number" or type(item) ~= "string" then return end

	if IsShiftKeyDown() then
		table.remove(addon.db.profile.loot, index)
	elseif IsControlKeyDown() then
		currentNoteItem = tonumber(select(3, item:find("item:(%d+):")))
		if not StaticPopupDialogs["NRTItemNoteDialog"] then
			StaticPopupDialogs["NRTItemNoteDialog"] = {
				text = nil,
				button1 = L["Set"],
				button2 = L["Cancel"],
				whileDead = 1,
				hideOnEscape = 1,
				timeout = 0,
				OnShow = function()
					-- We have to do this onshow to reset the previous text
					_G[this:GetName().."EditBox"]:SetText(addon.db.profile.itemNotes[currentNoteItem] or "")
				end,
				OnHide = function()
					_G[this:GetName().."EditBox"]:SetText("")
				end,
				EditBoxOnEnterPressed = function()
					local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
					addon.db.profile.itemNotes[currentNoteItem] = input
					this:GetParent():Hide()
				end,
				EditBoxOnEscapePressed = function()
					this:GetParent():Hide()
				end,
				OnAccept = function()
					local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
					addon.db.profile.itemNotes[currentNoteItem] = input
				end,
				hasEditBox = 1,
			}
		end
		StaticPopupDialogs["NRTItemNoteDialog"].text = itemNoteText:format(item)
		StaticPopup_Show("NRTItemNoteDialog")
	else
		SetItemRef(item)
	end
end

function addon:ShowTrackerForRaidIndex(index)
	createTrackerFrame()

	local raid = self.db.profile.tracked[index]
	if not raid then return end

	local formatter = formatters[self.db.profile.formatter]
	if not formatter then
		self:Print(L["The formatter %s isn't registered."]:format(self.db.profile.formatter))
		return
	end
	local text = formatter(raid, self.db.profile.loot, killedBosses)
	if not text then
		self:Print(L["The formatter %s didn't return any text for the raid ID %d."]:format(self.db.profile.formatter, index))
		return
	end

	setText(text)
	setFrameHeader(raid.date .. " " .. raid.name)
	trackerFrame.trackingId = index
	trackerFrame:Show()
end

local currentTrack = nil
local trackNoteText = L["Set note for track ID %d."]
local function trackerClickFunc(trackId)
	if type(trackId) ~= "number" then return end

	if IsShiftKeyDown() then
		if addon.db.profile.tracked[trackId] then
			table.remove(addon.db.profile.tracked, trackId)
		end
	elseif IsControlKeyDown() and addon.db.profile.tracked[trackId] then
		currentTrack = trackId
		if not StaticPopupDialogs["NRTTrackedNoteDialog"] then
			StaticPopupDialogs["NRTTrackedNoteDialog"] = {
				text = nil,
				button1 = L["Set"],
				button2 = L["Cancel"],
				whileDead = 1,
				hideOnEscape = 1,
				timeout = 0,
				OnShow = function()
					-- We have to do this onshow to reset the previous text
					_G[this:GetName().."EditBox"]:SetText(addon.db.profile.tracked[currentTrack].note or "")
				end,
				OnHide = function()
					_G[this:GetName().."EditBox"]:SetText("")
				end,
				EditBoxOnEnterPressed = function()
					local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
					addon.db.profile.tracked[currentTrack].note = input
					this:GetParent():Hide()
				end,
				EditBoxOnEscapePressed = function()
					this:GetParent():Hide()
				end,
				OnAccept = function()
					local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
					addon.db.profile.tracked[currentTrack].note = input
				end,
				hasEditBox = 1,
			}
		end
		StaticPopupDialogs["NRTTrackedNoteDialog"].text = trackNoteText:format(currentTrack)
		StaticPopup_Show("NRTTrackedNoteDialog")
	else
		-- Show the text area with the details from this track
		addon:ShowTrackerForRaidIndex(trackId)
	end
end

function addon:OnTooltipUpdate()
	local cat
	if #self.db.profile.tracked > 0 then
		local cat = tablet:AddCategory(
			"columns", 3,
			"text", L["Date"],
			"text2", L["Attendants"],
			"text3", L["Tracked"],
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text3R", 1,
			"child_text3G", 1,
			"child_text3B", 1,
			"showWithoutChildren", false,
			"child_func", trackerClickFunc
		)
		for i, v in ipairs(self.db.profile.tracked) do
			cat:AddLine(
				"text", date("%d/%m %H:%M", v.date),
				"text2", #v.attendants,
				"text3", v.name,
				"arg1", i
			)
		end
	end
	cat = tablet:AddCategory("columns", 1)
	cat:AddLine(
		"text", L["Click here to take raid attendance now."],
		"func", self.TakeAttendance,
		"arg1", self
	)
	if #self.db.profile.loot > 0 then
		cat = tablet:AddCategory(
			"columns", 4,
			"text", L["Date"],
			"text2", L["Player"],
			"text3", L["Item"],
			"text4", L["From"],
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text2R", 1,
			"child_text2G", 1,
			"child_text2B", 1,
			"showWithoutChildren", false,
			"child_func", itemClickFunc
		)
		for i, v in ipairs(self.db.profile.loot) do
			cat:AddLine(
				"text", date("%d/%m %H:%M", v.date),
				"text2", coloredName[v.player],
				"text3", v.item,
				"text4", v.from,
				"arg1", v.item,
				"arg2", i
			)
		end
	end

	tablet:SetHint(L["|cffeda55fClick|r an item to show. |cffeda55fShift-Click|r to remove. |cffeda55fCtrl-Click|r to edit notes."])
end

-----------------------------------------------------------------------
-- Formatters

function addon:RegisterOutputFormatter(name, formatter)
	if type(name) ~= "string" or type(formatter) ~= "function" then
		self:Print(L["Tried registering a formatter without a proper name or function."])
		return
	end
	if formatters[name] then
		self:Print(L["%s has already been registered."]:format(name))
		return
	end
	table.insert(outputFormatters, name)
	formatters[name] = formatter
end

-- Default formatter
addon:RegisterOutputFormatter("Nihilum", function(attendance, loot, killedBosses)
	return table.concat(attendance.attendants, "\n")
end)

addon:RegisterOutputFormatter("Example XML", function(attendance, loot, killedBosses)
	local lootByTracked = {}
	for i, v in ipairs(loot) do
		if not lootByTracked[v.from] then lootByTracked[v.from] = {} end
		table.insert(lootByTracked[v.from], v)
	end

	local text = "<raid zone=\"" .. attendance.zone .. "\">\n  <date>" .. date("%d/%m %H:%M", attendance.date) .. "</date>\n  "
	for k, v in pairs(lootByTracked) do
		text = text .. "<tracked source=\"" .. k .. "\">"
		for i, item in ipairs(v) do
			local itemName, _, itemRarity = GetItemInfo(item.item)
			local id = item.item:match("^|%x+|Hitem:(%d+):")
			text = text .. "\n    <item date=\"" .. date("%d/%m %H:%M", item.date) .. "\" player=\"" .. item.player .. "\" name=\"" .. (itemName or "?") .. "\" quality=\"" .. (itemRarity or "?") .. "\" id=\"" .. id .. "\" />"
		end
		text = text .. "\n  </tracked>\n  "
	end
	text = text .. "<players>"
	for i, v in ipairs(attendance.attendants) do
		text = text .. "\n    <player>" .. v .. "</player>"
	end
	text = text .. "\n  </players>\n  <bosses>"
	for i, v in ipairs(killedBosses) do
		text = text .. "\n    <boss>" .. v .. "</boss>"
	end
	text = text .. "\n  </bosses>\n</raid>"
	return text
end)

local ctrtDateFormat = "%m/%d/%y %H:%M:%S"
addon:RegisterOutputFormatter("CT Raidtracker XML", function(attendance, loot, killedBosses)
	local lootByTracked = {}
	for i, v in ipairs(loot) do
		if not lootByTracked[v.from] then lootByTracked[v.from] = {} end
		table.insert(lootByTracked[v.from], v)
	end

	local text = "<RaidInfo>"
	local players = ""

	local fdate = date(ctrtDateFormat, attendance.date)

	text = text .. "<key>" .. fdate .. "</key>"
	text = text .. "<start>" .. fdate .. "</start>"
	text = text .. "<end>" .. fdate .. "</end>"
	text = text .. "<zone>" .. attendance.zone .. "</zone>"

	text = text .. "<PlayerInfos>"

	for i, v in ipairs(attendance.attendants) do
		text = text .. "<key" .. i .. "><name>" .. v .. "</name></key" .. i .. ">"
		players = players .. "<key" .. i .. "><player>" .. v .. "</player><time>" .. fdate .. "</time></key" .. i .. ">"
	end

	text = text .. "</PlayerInfos>"

	text = text .. "<BossKills>"
	for i, v in ipairs(killedBosses) do
		text = text .. "<key" .. i .. "><name>" .. v .. "</name><time>" .. fdate .. "</time><attendees/></key" .. i .. ">"
	end
	text = text .. "</BossKills>"

	text = text .. "<note><![CDATA[" .. attendance.zone .. " - " .. attendance.name .. "]]></note>"
	text = text .. "<Join>" .. players .. "</Join>"
	text = text .. "<Leave>" .. players .. "</Leave>"

	text = text .. "<Loot>"

	for k, v in pairs(lootByTracked) do
		if k == attendance.name then
			for i, item in ipairs(v) do
				local itemName = GetItemInfo(item.item)
				local id = item.item:match("^|%x+|Hitem:(.*)")
				id = string.sub(id, 1, string.find(id, "|") - 1)

				local color = string.sub(item.item:match("^|(%x+)|"), 2)

				text = text .. "<key" .. i .. ">"
				text = text .. "<ItemName>" .. (itemName or "?") .. "</ItemName>"
				text = text .. "<ItemID>" .. id .. "</ItemID>"
				text = text .. "<Icon></Icon>"
				text = text .. "<Class></Class>"
				text = text .. "<SubClass></SubClass>"
				text = text .. "<Color>" .. color .. "</Color>"
				text = text .. "<Count>1</Count>"
				text = text .. "<Player>" .. item.player .. "</Player>"
				text = text .. "<Time>" ..  date(ctrtDateFormat, item.date) .. "</Time>"
				text = text .. "<Zone></Zone>"
				text = text .. "<Boss>" .. item.from .. "</Boss>"
				text = text .. "<Note></Note>"
				text = text .. "</key" .. i .. ">"
			end
		end
	end

	text = text .. "</Loot>"
	text = text .. "</RaidInfo>"
	return text
end)

